/*
 *  @(#) $Id: snom-example.c 12579 2011-07-11 15:14:11Z yeti-dn $
 *  Copyright (C) 2003,2004,2008 David Necas (Yeti)
 *  E-mail: yeti@gwyddion.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <gtk/gtk.h>
#include <libgwyddion/gwymacros.h>
#include <libgwyddion/gwymath.h>
#include <libgwymodule/gwymodule.h>
#include <libprocess/stats.h>
#include <libprocess/filters.h>
#include <libgwydgets/gwydgets.h>
#include <app/gwyapp.h>
#include <stdio.h>
#include "config.h"

#define SNOM_MOD_NAME PACKAGE_NAME

#define LIGHT_SPEED 299792458.0


#define SNOM_RUN_MODES (GWY_RUN_IMMEDIATE | GWY_RUN_INTERACTIVE)


/* Data for this function. */
typedef struct {
    gdouble xpos;
    gdouble ypos;
    gdouble zpos;
    gdouble aperture;
    gdouble wavelength;
    gdouble theta;
    gdouble phi;

    //gdouble fractile;
    //gsize mode;
    /* interface only */
    //gdouble absolute;
    gdouble min;
    gdouble max;
    gint xres;
    gint yres;
    const gchar *selected_filename;

    //gsize *fmap;
} SnomArgs;

typedef struct {
    GtkObject *xpos;
    GtkObject *ypos;
    GtkObject *zpos;
    GtkObject *aperture;
    GtkObject *wavelength;
    GtkObject *theta;
    GtkObject *phi;
    GtkWidget *filename;
    GtkWidget *filebutton;
    GtkWidget *file_selector;

    //GtkObject *fractile;
    //GSList *mode;
    /* interface only */
    //GtkObject *absolute;
    GwySIValueFormat *format;
    gboolean in_update;
    SnomArgs *args;
} SnomControls;

static gboolean    module_register                (void);
static gboolean    snom                      (GwyContainer *data,
                                                   GwyRunType run);
static gboolean    snom_dialog               (SnomArgs *args,
                                                   GwyDataField *dfield);
static void        snom_do                   (GwyContainer *data,
                                                   GwyDataField *dfield,
                                                   SnomArgs *args);
static void        zpos_changed               (GtkAdjustment *adj,
                                                   SnomControls *controls);
static void        xpos_changed               (GtkAdjustment *adj,
                                                   SnomControls *controls);
static void        theta_changed               (GtkAdjustment *adj,
                                                   SnomControls *controls);
static void        phi_changed               (GtkAdjustment *adj,
                                                   SnomControls *controls);
static void        ypos_changed               (GtkAdjustment *adj,
                                                   SnomControls *controls);
static void        aperture_changed          (GtkAdjustment *adj,
                                                   SnomControls *controls);
static void        wavelength_changed        (GtkAdjustment *adj,
                                                   SnomControls *controls);
static void        snom_load_args            (GwyContainer *container,
                                                   SnomArgs *args);
static void        snom_save_args            (GwyContainer *container,
                                                   SnomArgs *args);
static void        snom_dialog_update        (SnomControls *controls);

SnomArgs snom_defaults = {
    50,
    50,
    10e-9,
    50e-9,
    500e-9,
    0.7,
    0.7,
    0,
    0,
    0,
    0,
    NULL,
};

/* The module info. */
static GwyModuleInfo module_info = {
    GWY_MODULE_ABI_VERSION,
    &module_register,
    "Create data for FDTD modeling of a simple SNOM experiment",
    "Petr Klapetek <klapetek@gwyddion.net>",
    PACKAGE_VERSION,
    "David Nečas (Yeti) & Petr Klapetek",
    "2012",
};

/* This is the ONLY exported symbol.  The argument is the module info.
 * NO semicolon after. */
GWY_MODULE_QUERY(module_info)

/* Module registering function.
 * Called at Gwyddion startup and registeres one or more function.
 */
static gboolean
module_register(void)
{
    gwy_process_func_register(SNOM_MOD_NAME,
                              (GwyProcessFunc)&snom,
                              N_("/_Modeling/_SNOM..."),
                              NULL,
                              SNOM_RUN_MODES,
                              GWY_MENU_FLAG_DATA,
                              N_("Snom data."));

    return TRUE;
}

static gboolean
snom(GwyContainer *data, GwyRunType run)
{
    GwyDataField *dfield;
    GQuark dquark;
    SnomArgs args;
    gboolean ok;

    g_return_val_if_fail(run & SNOM_RUN_MODES, FALSE);
    /* Obtain the current data and mask fields */
    gwy_app_data_browser_get_current(GWY_APP_DATA_FIELD_KEY, &dquark,
                                     GWY_APP_DATA_FIELD, &dfield,
                                     NULL);
    /* Set up parameters */
    args = snom_defaults;
    snom_load_args(gwy_app_settings_get(), &args);
    args.xres = gwy_data_field_get_xres(dfield);
    args.yres = gwy_data_field_get_yres(dfield);
    args.max = MAX(100e-9, gwy_data_field_get_max(dfield)*1000); //give some boundary to maximum distance of the probe

    /* Possibly present the GUI */
    ok = (run != GWY_RUN_INTERACTIVE) || snom_dialog(&args, dfield);
    if (ok) {
        snom_do(data, dfield, &args);
        snom_save_args(gwy_app_settings_get(), &args);
    }

    return TRUE;
}


static gboolean
is_in_cylinder(gint i, gint j, gint k, gint ic, gint jc, gint kfrom, gint kto, gdouble radius)
{
    if (k>kfrom && k<kto && ((i-ic)*(i-ic)+(j-jc)*(j-jc))<=(radius*radius)) return 1;
    else return 0;
}



static void
snom_do(GwyContainer *data,
             GwyDataField *dfield,
             SnomArgs *args)
{
    GtkWidget *dialog;
    gint i, j, k, xres, yres, zres, errno = 0;
    FILE *fw;
    gchar *filename;
    gfloat ***epsilon;
    gfloat ***sigma;
    gfloat mu = 1;
    gfloat sigast = 0;
    gdouble wavefreq, width;
    gdouble xreal, yreal, xpos, ypos, zval, zmin, zmax, diff, dt, val, ex;
    gint izmin, izmax, izval, iheight;
    gint nsteps;
    gdouble radius, ffx, ffy, ffz;

    xres = 200;
    yres = 200;
    zres = 200;

    printf("Exporting for position %d %d %g, wavelength %g, aperture %g, theta %g, phi %g to directory %s\n", 
           (gint)args->xpos, (gint)args->ypos, args->zpos, args->wavelength, args->aperture, args->theta, args->phi, 
           args->selected_filename);



    filename = g_build_filename(args->selected_filename, "snom_material.bin", NULL);
    fw = fopen(filename, "wb");
    if (!fw) {
        dialog = gtk_message_dialog_new (NULL,
                                         GTK_DIALOG_DESTROY_WITH_PARENT,
                                         GTK_MESSAGE_ERROR,
                                         GTK_BUTTONS_CLOSE,
                                         "Error creating file '%s':%s",
                                         filename, g_strerror (errno));
        gtk_dialog_run (GTK_DIALOG (dialog));
        gtk_widget_destroy (dialog);
    }

    epsilon = (gfloat ***) g_malloc(xres*sizeof(gfloat**));
    for (i=0; i<xres; i++)
    {
        epsilon[i] = (gfloat **) g_malloc(yres*sizeof(gfloat*));
        for (j=0; j<yres; j++)
        {
            epsilon[i][j] = (gfloat *) g_malloc(zres*sizeof(gfloat));
        }
    }

    sigma = (gfloat ***) g_malloc(xres*sizeof(gfloat**));
    for (i=0; i<xres; i++)
    {
        sigma[i] = (gfloat **) g_malloc(yres*sizeof(gfloat*));
        for (j=0; j<yres; j++)
        {
            sigma[i][j] = (gfloat *) g_malloc(zres*sizeof(gfloat));
        }
    }

    /*create the material model*/

    diff = args->wavelength/20;
    zmin = gwy_data_field_get_min(dfield); /*surface minimum in real coordinates*/
    zmax = gwy_data_field_get_max(dfield); /*surface maximum in real coordinates*/
    xreal = gwy_data_field_get_xreal(dfield);
    yreal = gwy_data_field_get_yreal(dfield);

    izmin = zmin/diff; /*surface minimum in pixel coordinates*/
    izmax = zmax/diff; /*surface maximum in pixel coordinates*/

    iheight = 30+gwy_data_field_get_dval_real(dfield, args->xpos, args->ypos, GWY_INTERPOLATION_BILINEAR)+args->zpos/diff; /*source position*/

    for (i=0; i<xres; i++)
    {
        for (j=0; j<yres; j++)
        {
            for (k=0; k<zres; k++) {
                epsilon[i][j][k] = 1;
                sigma[i][j][k] = 0;

                /*surface*/
                xpos = MAX(0, gwy_data_field_itor(dfield, args->xpos)  - diff*(i - xres/2));
                ypos = MAX(0, gwy_data_field_jtor(dfield, args->ypos)  - diff*(j - yres/2));
                zval = gwy_data_field_get_dval_real(dfield, xpos, ypos, GWY_INTERPOLATION_BILINEAR);
                izval = 30+(gint)(zval/diff) - izmin;
                //printf("%g %g %g %d\n", xpos, ypos, zval, izval);
                
                if (k<=izval) {
                    epsilon[i][j][k] = 4;
                    sigma[i][j][k] = 100000;
                }


                /*probe, for beginning use only point source*/
                /*if (is_in_cylinder(i, j, k, xres/2, yres/2, iheight, iheight+40, 10))
                {
                    epsilon[i][j][k] = 4;
                    sigma[i][j][k] = 100000;
                }
                if (is_in_cylinder(i, j, k, xres/2, yres/2, iheight, iheight+35, 6))
                {
                    epsilon[i][j][k] = 1;
                    sigma[i][j][k] = 0;
                }*/



            }
        }
    }

   
    /*save the files for the material model*/
    fwrite(&xres, sizeof(gint), 1, fw);
    fwrite(&yres, sizeof(gint), 1, fw);
    fwrite(&zres, sizeof(gint), 1, fw);
    for (i=0; i<xres; i++)
    {
        for (j=0; j<yres; j++)
        {
                fwrite(epsilon[i][j], sizeof(gfloat), zres, fw);
        }
    }
    for (i=0; i<xres; i++)
    {
        for (j=0; j<yres; j++)
        {
                fwrite(sigma[i][j], sizeof(gfloat), zres, fw);
        }
    }

    for (i=0; i<xres; i++)
    {
        for (j=0; j<yres; j++)
        {
            for (k=0; k<zres; k++) {
                fwrite(&mu, sizeof(gfloat), 1, fw);
            }
        }
    }
    for (i=0; i<xres; i++)
    {
        for (j=0; j<yres; j++)
        {
            for (k=0; k<zres; k++) {
                fwrite(&sigast, sizeof(gfloat), 1, fw);
            }
        }
    }

    fclose(fw);


    /*create computation parameters file*/
    dt = 1.0/LIGHT_SPEED/sqrt(3.0/diff/diff);
    nsteps = 300;


    filename = g_build_filename(args->selected_filename, "snom.par", NULL);
    fw = fopen(filename, "w");
    if (!fw) {
        dialog = gtk_message_dialog_new (NULL,
                                         GTK_DIALOG_DESTROY_WITH_PARENT,
                                         GTK_MESSAGE_ERROR,
                                         GTK_BUTTONS_CLOSE,
                                         "Error creating file '%s':%s",
                                         filename, g_strerror (errno));
        gtk_dialog_run (GTK_DIALOG (dialog));
        gtk_widget_destroy (dialog);
    }


    //source je ve vysce iheight a uprostred pole. bude odstinen primitivne.
    //nfff je jakoby se nechumelilo
    //boundary je liao

    radius = 10000;
    ffx = radius*cos(args->theta)*sin(args->phi);
    ffy = radius*sin(args->theta)*sin(args->phi);
    ffz = radius*cos(args->phi);

    fprintf(fw, "POOL\n%d %d %d %g %g %g\n\n", xres, yres, zres, diff, diff, diff);
    fprintf(fw, "COMP\n%d\n\n", 100/*nsteps*/);
    fprintf(fw, "VERBOSE\n%d\n\n", 3);
    fprintf(fw, "OUT_IMAGE\nEpsilon 1000 %d %d %d epsilon_xplane nologscale\n\n", xres/2, -1, -1);
    fprintf(fw, "OUT_IMAGE\nEpsilon 1000 %d %d %d epsilon_yplane nologscale\n\n", -1, yres/2, -1);
    fprintf(fw, "OUT_IMAGE\nEpsilon 1000 %d %d %d epsilon_zplane nologscale\n\n", -1, -1, zres/2);
    fprintf(fw, "OUT_IMAGE\nSigma 1000 %d %d %d sigma_xplane nologscale\n\n", xres/2, -1, -1);
    fprintf(fw, "OUT_IMAGE\nSigma 1000 %d %d %d sigma_yplane nologscale\n\n", -1, yres/2, -1);
    fprintf(fw, "OUT_IMAGE\nSigma 1000 %d %d %d sigma_zplane nologscale\n\n", -1, -1, zres/2);
    fprintf(fw, "OUT_IMAGE\nEx 20 %d %d %d ex_xplane nologscale\n\n", xres/2, -1, -1);
    fprintf(fw, "OUT_IMAGE\nEx 20 %d %d %d ex_yplane nologscale\n\n", -1, yres/2, -1);
    fprintf(fw, "OUT_IMAGE\nEx 20 %d %d %d ex_zplane nologscale\n\n", -1, -1, zres/2);
    fprintf(fw, "MEDIUM_LINEAR\nsnom_material.bin\n\n");
    fprintf(fw, "BOUNDARY_ALL\nliao\n\n");
    fprintf(fw, "SOURCE_POINT\n%d %d %d snom_source.txt\n\n", xres/2, yres/2, iheight);
    fprintf(fw, "NFFF\n%d %d %d %d %d %d 1 10000000\n\n", 20, 20, 20, xres-20, yres-20, zres-20);
    fprintf(fw, "NFFF_RAMAHI_POINT\n%d %d %d ffpnt\n\n", (gint)ffx, (gint)ffy, (gint)ffz);

    fclose(fw);
    

    /*create source*/
    wavefreq =LIGHT_SPEED*2*M_PI/args->wavelength;
    width = 20*dt;


    filename = g_build_filename(args->selected_filename, "snom_source.txt", NULL);
    fw = fopen(filename, "w");
    if (!fw) {
        dialog = gtk_message_dialog_new (NULL,
                                         GTK_DIALOG_DESTROY_WITH_PARENT,
                                         GTK_MESSAGE_ERROR,
                                         GTK_BUTTONS_CLOSE,
                                         "Error creating file '%s':%s",
                                         filename, g_strerror (errno));
        gtk_dialog_run (GTK_DIALOG (dialog));
        gtk_widget_destroy (dialog);
    }


    printf("%g %g %g\n", wavefreq, dt, width);
    fprintf(fw, "%d\n", nsteps);
    for (i=0; i<nsteps; i++)
    {
        val = 5*sin(wavefreq*dt*(double)i);
        if (i<100) val *= 0.01*((double)(i));

        ex = ((double)i*dt - 4*width)/sqrt(2)/width;
        val *= exp(-ex*ex);
        fprintf(fw, "%d %g %g %g %g %g %g\n", i, val, 0.0, 0.0, 0.0, 0.0, 0.0);

    }

    fclose(fw);
}



gboolean            
dir_clicked(GtkWidget *widget, SnomControls *controls)
{
    GtkWidget *dialog;
    dialog = gtk_file_chooser_dialog_new ("Select folder",
                                          GTK_WINDOW(gtk_widget_get_toplevel (widget)),
                                          GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER,
                                          GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                          GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
                                          NULL);
    if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
    {
        controls->args->selected_filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
        gtk_button_set_label(GTK_BUTTON(controls->filename), controls->args->selected_filename);
    }
    gtk_widget_destroy (dialog);
   
   return TRUE;
} 

static gboolean
snom_dialog(SnomArgs *args,
                 GwyDataField *dfield)
{
    GtkWidget *dialog, *table, *spin, *widget;
    GSList *group;
    GwySIValueFormat *fmt;
    SnomControls controls;
    enum { RESPONSE_RESET = 1 };
    gint response;

    controls.args = args;
    controls.args->selected_filename = NULL;

    dialog = gtk_dialog_new_with_buttons(_("Snom"), NULL, 0,
                                         _("Reset"), RESPONSE_RESET,
                                         GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                         GTK_STOCK_OK, GTK_RESPONSE_OK,
                                         NULL);
    gtk_dialog_set_has_separator(GTK_DIALOG(dialog), FALSE);
    gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_OK);

    table = gtk_table_new(11, 3, FALSE);
    gtk_table_set_row_spacings(GTK_TABLE(table), 2);
    gtk_table_set_col_spacings(GTK_TABLE(table), 6);
    gtk_container_set_border_width(GTK_CONTAINER(table), 4);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), table,
                       FALSE, FALSE, 4);


    controls.xpos = gtk_adjustment_new(args->xpos, 0, args->xres, 1, 10, 0);
    spin = gwy_table_attach_spinbutton(table, 1, _("X position:"),
                                       "px",
                                       controls.xpos);
    gtk_spin_button_set_digits(GTK_SPIN_BUTTON(spin), 0);
    g_signal_connect(controls.xpos, "value_changed",
                     G_CALLBACK(xpos_changed), &controls);

    controls.ypos = gtk_adjustment_new(args->ypos, 0, args->yres, 1, 10, 0);
    spin = gwy_table_attach_spinbutton(table, 2, _("X position:"),
                                       "px",
                                       controls.ypos);
    gtk_spin_button_set_digits(GTK_SPIN_BUTTON(spin), 0);
    g_signal_connect(controls.ypos, "value_changed",
                     G_CALLBACK(ypos_changed), &controls);


    fmt = gwy_data_field_get_value_format_z(dfield,
                                            GWY_SI_UNIT_FORMAT_VFMARKUP, NULL);
    controls.format = fmt;
    controls.zpos = gtk_adjustment_new(args->zpos/fmt->magnitude,
                                           args->min/fmt->magnitude,
                                           args->max/fmt->magnitude,
                                           MIN(10e-9, (args->max - args->min)/100)
                                                /fmt->magnitude,
                                           MIN(10e-9, (args->max - args->min)/10)
                                                /fmt->magnitude,
                                           0);
    spin = gwy_table_attach_spinbutton(table, 3, _("Z position:"),
                                       fmt->units,
                                       controls.zpos);
    gtk_spin_button_set_digits(GTK_SPIN_BUTTON(spin), fmt->precision);
    g_signal_connect(controls.zpos, "value_changed",
                     G_CALLBACK(zpos_changed), &controls);

    controls.aperture = gtk_adjustment_new(args->aperture/fmt->magnitude,
                                           args->min/fmt->magnitude,
                                           args->max/fmt->magnitude,
                                           MIN(10e-9, (args->max - args->min)/100)
                                                /fmt->magnitude,
                                           MIN(10e-9, (args->max - args->min)/10)
                                                /fmt->magnitude,
                                           0);
    spin = gwy_table_attach_spinbutton(table, 4, _("Aperture size:"),
                                       fmt->units,
                                       controls.aperture);
    gtk_spin_button_set_digits(GTK_SPIN_BUTTON(spin), fmt->precision);
    g_signal_connect(controls.aperture, "value_changed",
                     G_CALLBACK(aperture_changed), &controls);

    controls.wavelength = gtk_adjustment_new(args->wavelength/fmt->magnitude,
                                           args->min/fmt->magnitude,
                                           args->max/fmt->magnitude,
                                           MIN(1e-9, (args->max - args->min)/100)
                                                /fmt->magnitude,
                                           MIN(1e-9, (args->max - args->min)/10)
                                                /fmt->magnitude,
                                           0);
    spin = gwy_table_attach_spinbutton(table, 5, _("Wavelength:"),
                                       fmt->units,
                                       controls.wavelength);
    gtk_spin_button_set_digits(GTK_SPIN_BUTTON(spin), fmt->precision);
    g_signal_connect(controls.wavelength, "value_changed",
                     G_CALLBACK(wavelength_changed), &controls);

    controls.theta = gtk_adjustment_new(args->theta, 0, 2*G_PI, 0.1, 1, 0);
    spin = gwy_table_attach_spinbutton(table, 6, _("Far field θ:"),
                                       "rad",
                                       controls.theta);
    gtk_spin_button_set_digits(GTK_SPIN_BUTTON(spin), 2);
    g_signal_connect(controls.theta, "value_changed",
                     G_CALLBACK(theta_changed), &controls);

    controls.phi = gtk_adjustment_new(args->phi, 0, G_PI/2.0, 1, 10, 0);
    spin = gwy_table_attach_spinbutton(table, 7, _("Far field φ:"),
                                       "rad",
                                       controls.phi);
    gtk_spin_button_set_digits(GTK_SPIN_BUTTON(spin), 2);
    g_signal_connect(controls.phi, "value_changed",
                     G_CALLBACK(phi_changed), &controls);


    widget = gtk_label_new("Working directory:");
    gtk_misc_set_alignment(GTK_MISC(widget), 0.0, 0.5);
    gtk_table_attach_defaults(GTK_TABLE(table), widget, 0, 2, 8, 9);

    controls.filename = gtk_button_new_with_label("Click to select...");
    g_signal_connect(controls.filename, "clicked", 
                     G_CALLBACK(dir_clicked), &controls);
    gtk_table_attach_defaults(GTK_TABLE(table), controls.filename, 0, 3, 9, 10);


    controls.in_update = FALSE;

    gtk_widget_show_all(dialog);
    do {
        response = gtk_dialog_run(GTK_DIALOG(dialog));
        switch (response) {
            case GTK_RESPONSE_CANCEL:
            case GTK_RESPONSE_DELETE_EVENT:
            g_free(controls.format);
            gtk_widget_destroy(dialog);
            case GTK_RESPONSE_NONE:
            return FALSE;
            break;

            case GTK_RESPONSE_OK:
            break;

            case RESPONSE_RESET:
            args->zpos = snom_defaults.zpos;
            snom_dialog_update(&controls);
            break;

            default:
            g_assert_not_reached();
            break;
        }
    } while (response != GTK_RESPONSE_OK);

    g_free(controls.format);
    gtk_widget_destroy(dialog);

    return TRUE;
}




static void
xpos_changed(GtkAdjustment *adj,
                 SnomControls *controls)
{
    if (controls->in_update)
        return;

    controls->in_update = TRUE;
    controls->args->xpos = gtk_adjustment_get_value(adj);
    snom_dialog_update(controls);
    controls->in_update = FALSE;
}

static void
ypos_changed(GtkAdjustment *adj,
                 SnomControls *controls)
{
    if (controls->in_update)
        return;

    controls->in_update = TRUE;
    controls->args->ypos = gtk_adjustment_get_value(adj);
    snom_dialog_update(controls);
    controls->in_update = FALSE;
}

static void
zpos_changed(GtkAdjustment *adj,
                 SnomControls *controls)
{
    if (controls->in_update)
        return;

    controls->in_update = TRUE;
    controls->args->zpos = gtk_adjustment_get_value(adj)
                               *controls->format->magnitude;
    snom_dialog_update(controls);
    controls->in_update = FALSE;
}

static void
aperture_changed(GtkAdjustment *adj,
                 SnomControls *controls)
{
    if (controls->in_update)
        return;

    controls->in_update = TRUE;
    controls->args->aperture = gtk_adjustment_get_value(adj)
                               *controls->format->magnitude;
    snom_dialog_update(controls);
    controls->in_update = FALSE;
}

static void
wavelength_changed(GtkAdjustment *adj,
                 SnomControls *controls)
{
    if (controls->in_update)
        return;

    controls->in_update = TRUE;
    controls->args->wavelength = gtk_adjustment_get_value(adj)
                               *controls->format->magnitude;
    snom_dialog_update(controls);
    controls->in_update = FALSE;
}

static void
theta_changed(GtkAdjustment *adj,
                 SnomControls *controls)
{
    if (controls->in_update)
        return;

    controls->in_update = TRUE;
    controls->args->theta = gtk_adjustment_get_value(adj);
    snom_dialog_update(controls);
    controls->in_update = FALSE;
}

static void
phi_changed(GtkAdjustment *adj,
                 SnomControls *controls)
{
    if (controls->in_update)
        return;

    controls->in_update = TRUE;
    controls->args->phi = gtk_adjustment_get_value(adj);
    snom_dialog_update(controls);
    controls->in_update = FALSE;
}

static const gchar xpos_key[] = "/module/" SNOM_MOD_NAME "/xpos";
static const gchar ypos_key[] = "/module/" SNOM_MOD_NAME "/ypos";
static const gchar zpos_key[] = "/module/" SNOM_MOD_NAME "/zpos";
static const gchar aperture_key[] = "/module/" SNOM_MOD_NAME "/aperture";
static const gchar wavelength_key[] = "/module/" SNOM_MOD_NAME "/wavelength";
static const gchar theta_key[] = "/module/" SNOM_MOD_NAME "/theta";
static const gchar phi_key[] = "/module/" SNOM_MOD_NAME "/phi";

static void
snom_load_args(GwyContainer *container,
                    SnomArgs *args)
{
    *args = snom_defaults;

    if (gwy_container_contains_by_name(container, xpos_key))
        args->xpos = gwy_container_get_double_by_name(container,
                                                          xpos_key);
    args->xpos = CLAMP(args->xpos, 0.0, 10000);

    if (gwy_container_contains_by_name(container, ypos_key))
        args->ypos = gwy_container_get_double_by_name(container,
                                                          ypos_key);
    args->ypos = CLAMP(args->ypos, 0.0, 10000);

    if (gwy_container_contains_by_name(container, zpos_key))
        args->zpos = gwy_container_get_double_by_name(container,
                                                          zpos_key);
    args->zpos = CLAMP(args->zpos, 0.0, 1000);

    if (gwy_container_contains_by_name(container, aperture_key))
        args->aperture = gwy_container_get_double_by_name(container,
                                                          aperture_key);
    args->aperture = CLAMP(args->aperture, 0.0, 1000);

    if (gwy_container_contains_by_name(container, wavelength_key))
        args->wavelength = gwy_container_get_double_by_name(container,
                                                          wavelength_key);
    args->wavelength = CLAMP(args->wavelength, 0.0, 1000);

    if (gwy_container_contains_by_name(container, theta_key))
        args->theta = gwy_container_get_double_by_name(container,
                                                          theta_key);
    args->theta = CLAMP(args->theta, 0.0, 2*G_PI);

    if (gwy_container_contains_by_name(container, phi_key))
        args->phi = gwy_container_get_double_by_name(container,
                                                          phi_key);
    args->phi = CLAMP(args->phi, 0.0, G_PI/2.0);
}

static void
snom_save_args(GwyContainer *container,
                SnomArgs *args)
{
    gwy_container_set_double_by_name(container, xpos_key, args->xpos);
    gwy_container_set_double_by_name(container, ypos_key, args->ypos);
    gwy_container_set_double_by_name(container, zpos_key, args->zpos);
    gwy_container_set_double_by_name(container, aperture_key, args->aperture);
    gwy_container_set_double_by_name(container, wavelength_key, args->wavelength);
    gwy_container_set_double_by_name(container, theta_key, args->theta);
    gwy_container_set_double_by_name(container, phi_key, args->phi);
}

static void
snom_dialog_update(SnomControls *controls)
{
    SnomArgs *args;
    GSList *l;

    args = controls->args;
    gtk_adjustment_set_value(GTK_ADJUSTMENT(controls->zpos), args->zpos/controls->format->magnitude);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(controls->aperture), args->aperture/controls->format->magnitude);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(controls->wavelength), args->wavelength/controls->format->magnitude);

    gtk_adjustment_set_value(GTK_ADJUSTMENT(controls->xpos), args->xpos);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(controls->ypos), args->ypos);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(controls->theta), args->theta);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(controls->phi), args->phi);


}


/* vim: set cin et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
